@@ -1,5 +1,6 @@
 import click
 import click_config_file
+import concurrent.futures
 import configparser
 import glob
 import json
@@ -151,21 +152,28 @@ class IndicesMappingsCollector(object):
 
 
 class IndicesStatsCollector(object):
-    def __init__(self, es_client, timeout, parse_indices=False, metrics=None, fields=None):
+    def __init__(self, es_client, timeout, parse_indices=False,
+                 indices=None, metrics=None, fields=None):
         self.metric_name_list = ['es', 'indices_stats']
         self.description = 'Indices Stats'
 
         self.es_client = es_client
         self.timeout = timeout
         self.parse_indices = parse_indices
+        self.indices = indices
         self.metrics = metrics
         self.fields = fields
 
     def collect(self):
         try:
-            response = self.es_client.indices.stats(metric=self.metrics, fields=self.fields, request_timeout=self.timeout)
+            response = self.es_client.indices.stats(index=self.indices,
+                                                    metric=self.metrics,
+                                                    fields=self.fields,
+                                                    request_timeout=self.timeout)
 
-            metrics = indices_stats_parser.parse_response(response, self.parse_indices, self.metric_name_list)
+            metrics = indices_stats_parser.parse_response(response,
+                                                          self.parse_indices,
+                                                          self.metric_name_list)
             metric_dict = group_metrics(metrics)
         except ConnectionTimeout:
             log.warning('Timeout while fetching %(description)s (timeout %(timeout_s)ss).',
@@ -339,6 +347,16 @@ INDICES_STATS_METRICS_OPTIONS = [
 ]
 
 
+def indices_stats_indices_parser(ctx, param, value):
+    if value is None:
+        return None
+
+    if value in ('*', '_all', ''):
+        return value
+    else:
+        return value.split(',')
+
+
 def indices_stats_fields_parser(ctx, param, value):
     if value is None:
         return None
@@ -349,6 +367,23 @@ def indices_stats_fields_parser(ctx, param, value):
         return value.split(',')
 
 
+def split_http_header(header_string):
+    """Splits a colon-separated string into header and value"""
+    parts = tuple(part.strip() for part in header_string.split(":", 1))
+    if len(parts) != 2 or any(part == '' for part in parts):
+        msg = "Invalid header '{}'. Use colon to separate name and value".format(header_string)
+        raise click.BadParameter(msg)
+    return parts
+
+
+def http_headers_parser(ctx, param, headers):
+    """Turn header list into a dict {header: value, ...}"""
+    if headers is None:
+        return None
+
+    return dict(split_http_header(header_string) for header_string in headers)
+
+
 def configparser_enum_conv(enum):
     lower_enums = tuple(e.lower() for e in enum)
 
@@ -394,6 +429,13 @@ CONFIGPARSER_CONVERTERS = {
 @click.option('--basic-password',
               help='Password for basic authentication with nodes. '
                    'Must be specified if "--basic-user" is provided.')
+@click.option('--header', '-H',
+              multiple=True,
+              callback=http_headers_parser,
+              help='HTTP header to include in requests to ElasticSearch cluster. '
+                   'Header name and value should be separated by colon, e.g. '
+                   '"Authorization: Bearer xxxxx". Several headers can be added '
+                   'by repeating the -H parameter.')
 @click.option('--port', '-p', default=9206,
               help='Port to serve the metrics endpoint on. (default: 9206)')
 @click.option('--query-disable', default=False, is_flag=True,
@@ -411,6 +453,9 @@ CONFIGPARSER_CONVERTERS = {
                    'in filename order. '
                    'Can be absolute, or relative to the current working directory. '
                    '(default: ./config)')
+@click.option('--threads', type=click.IntRange(min=1), default=1,
+              help='Enables concurrent query execution using the number of threads specified. '
+                   '(default: 1)')
 @click.option('--cluster-health-disable', default=False, is_flag=True,
               help='Disable cluster health monitoring.')
 @click.option('--cluster-health-timeout', default=10.0,
@@ -441,6 +486,11 @@ CONFIGPARSER_CONVERTERS = {
 @click.option('--indices-stats-mode', default='cluster',
               type=click.Choice(['cluster', 'indices']),
               help='Detail mode for indices stats monitoring. (default: cluster)')
+@click.option('--indices-stats-indices',
+              callback=indices_stats_indices_parser,
+              help='Limit indices stats to specific indices. '
+                   'Only takes effect if "--indices-stats-mode=indices". '
+                   'Indices should be separated by commas e.g. index1,index2.')
 @click.option('--indices-stats-metrics',
               type=MultiChoice(INDICES_STATS_METRICS_OPTIONS),
               help='Limit indices stats to specific metrics. '
@@ -448,7 +498,7 @@ CONFIGPARSER_CONVERTERS = {
 @click.option('--indices-stats-fields',
               callback=indices_stats_fields_parser,
               help='Include fielddata info for specific fields. '
-                   'Fields should be separated by commas e.g. indices,fs. '
+                   'Fields should be separated by commas e.g. field1,field2. '
                    'Use \'*\' for all.')
 @click.option('--json-logging', '-j', default=False, is_flag=True,
               help='Turn on json logging.')
@@ -461,22 +511,36 @@ CONFIGPARSER_CONVERTERS = {
 def cli(**options):
     """Export Elasticsearch query results to Prometheus."""
     if options['basic_user'] and options['basic_password'] is None:
-        click.BadOptionUsage('basic_user', 'Username provided with no password.')
+        raise click.BadOptionUsage('basic_user', 'Username provided with no password.')
     elif options['basic_user'] is None and options['basic_password']:
-        click.BadOptionUsage('basic_password', 'Password provided with no username.')
+        raise click.BadOptionUsage('basic_password', 'Password provided with no username.')
     elif options['basic_user']:
         http_auth = (options['basic_user'], options['basic_password'])
     else:
         http_auth = None
 
     if not options['ca_certs'] and options['client_cert']:
-        click.BadOptionUsage('client_cert', '--client-cert can only be used when --ca-certs is provided.')
+        raise click.BadOptionUsage('client_cert',
+                                   '--client-cert can only be used when --ca-certs is provided.')
     elif not options['ca_certs'] and options['client_key']:
-        click.BadOptionUsage('client_key', '--client-key can only be used when --ca-certs is provided.')
+        raise click.BadOptionUsage('client_key',
+                                   '--client-key can only be used when --ca-certs is provided.')
     elif options['client_cert'] and not options['client_key']:
-        click.BadOptionUsage('client_cert', '--client-key must be provided when --client-cert is used.')
+        raise click.BadOptionUsage('client_cert',
+                                   '--client-key must be provided when --client-cert is used.')
     elif not options['client_cert'] and options['client_key']:
-        click.BadOptionUsage('client_key', '--client-cert must be provided when --client-key is used.')
+        raise click.BadOptionUsage('client_key',
+                                   '--client-cert must be provided when --client-key is used.')
+
+    if options['indices_stats_indices'] and options['indices_stats_mode'] != 'indices':
+        raise click.BadOptionUsage('indices_stats_indices',
+                                   '--indices-stats-mode must be "indices" for '
+                                   '--indices-stats-indices to be used.')
+
+    executor = None
+    num_threads = options['threads']
+    if num_threads > 1:
+        executor = concurrent.futures.ThreadPoolExecutor(max_workers=num_threads)
 
     log_handler = logging.StreamHandler()
     log_format = '[%(asctime)s] %(name)s.%(levelname)s %(threadName)s %(message)s'
@@ -499,10 +563,12 @@ def cli(**options):
                                   ca_certs=options['ca_certs'],
                                   client_cert=options['client_cert'],
                                   client_key=options['client_key'],
+                                  headers=options['header'],
                                   http_auth=http_auth)
     else:
         es_client = Elasticsearch(es_cluster,
                                   verify_certs=False,
+                                  headers=options['header'],
                                   http_auth=http_auth)
 
     scheduler = None
@@ -540,7 +606,7 @@ def cli(**options):
         if queries:
             for query_name, (interval, timeout, indices, query,
                              on_error, on_missing) in queries.items():
-                schedule_job(scheduler, interval,
+                schedule_job(scheduler, executor, interval,
                              run_query, es_client, query_name, indices, query,
                              timeout, on_error, on_missing)
         else:
@@ -570,6 +636,7 @@ def cli(**options):
         REGISTRY.register(IndicesStatsCollector(es_client,
                                                 options['indices_stats_timeout'],
                                                 parse_indices=parse_indices,
+                                                indices=options['indices_stats_indices'],
                                                 metrics=options['indices_stats_metrics'],
                                                 fields=options['indices_stats_fields']))
 
